(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2016)
Author:  Sarah Winkler <sarah.winkler@uibk.ac.at> (2016, 2017)
License: LGPL (see file COPYING.LESSER)
*)
theory Ordered_Completion_Impl
  imports
    Ordered_Completion
    TRS.Trs_Impl
    TRS.Q_Restricted_Rewriting_Impl
    Ord.KBO_Impl
    Ord.Reduction_Order_Impl
    Ord.RPO
    Equational_Reasoning
begin

hide_const (open) Exact_Tree_Automata_Completion.fground

hide_const (open) Ramsey.choice
  
datatype ('f, 'v) oc_irule =
  OC_Deduce "('f, 'v) term" "('f, 'v) term" "('f, 'v) term"
  | OC_Orientl "('f, 'v) term" "('f, 'v) term"
  | OC_Orientr "('f, 'v) term" "('f, 'v) term"
  | OC_Delete "('f, 'v) term"
  | OC_Compose "('f, 'v) term" "('f, 'v) term" "('f, 'v) term"
  | OC_Simplifyl "('f, 'v) term" "('f, 'v) term" "('f, 'v) term"
  | OC_Simplifyr "('f, 'v) term" "('f, 'v) term" "('f, 'v) term"
  | OC_Collapse "('f, 'v) term" "('f, 'v) term" "('f, 'v) term"

definition "sym_list xs = List.union xs (map (\<lambda>(x, y). (y, x)) xs)"
  
lemma set_sym_list [simp]: "set (sym_list xs) = (set xs)\<^sup>\<leftrightarrow>" by (auto simp: sym_list_def)

definition check_rstep_p
  where
    "check_rstep_p c \<rho> p s t =
      (if p \<notin> set (poss_list t) then error (shows ''no step possible at this position'')
       else
         let (l, r) = \<rho> in
         (case match_list Var [(l, s |_ p), (r, t |_ p)] of
              None \<Rightarrow> error (shows ''rule does not match'')
            | Some \<sigma> \<Rightarrow>
              if t = replace_at s p (r \<cdot> \<sigma>) then c (l \<cdot> \<sigma>) (r \<cdot> \<sigma>)
              else error (shows ''result does not match'')))"

definition check_step_rule
  where
    "check_step_rule c \<rho> s t =
      check_exm (\<lambda>p. check_rstep_p c \<rho> p s t) (poss_list s) (\<lambda>_. shows '' is not a reduct'')"

lemma check_step_rule:
  assumes ok: "isOK (check_step_rule chk (l,r) s t)" 
  shows "\<exists>C \<sigma> p. s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<and> isOK (chk (l \<cdot> \<sigma>) (r \<cdot> \<sigma>))"
proof -
  from ok[unfolded check_step_rule_def case_prod_beta isOK_update_error isOK_existsM]
    obtain p where ok_p:"isOK (check_rstep_p chk (l,r) p s t)" "p \<in> poss s" by auto
  let ?p = "\<lambda> \<sigma>.(if t = replace_at s p (r \<cdot> \<sigma>) then chk (l \<cdot> \<sigma>) (r \<cdot> \<sigma>)
                 else (error (shows ''result does not match'')))"
  let ?ml = "match_list Var [(l, s |_ p), (r, t |_ p)]"
  from ok_p[unfolded check_rstep_p_def] obtain \<sigma> where \<sigma>:"?ml = Some \<sigma>" "isOK(?p \<sigma>)" 
    by (cases ?ml, auto)
  from match_list_sound[OF \<sigma>(1)] have at_p:"l \<cdot> \<sigma> = s |_ p" "r \<cdot> \<sigma> = t |_ p" by auto
  from \<sigma>(2) have rlcheck:"isOK(chk (l \<cdot> \<sigma>) (r \<cdot> \<sigma>))" and t:"t = (ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle>"
    unfolding split by auto
  let ?C = "ctxt_of_pos_term p s"
  from at_p ctxt_supt_id[OF ok_p(2)] t have s:"s = ?C\<langle>l \<cdot> \<sigma>\<rangle>"  and t:"t = ?C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  from s t rlcheck show ?thesis by force
qed

fun check_step
  where
    "check_step c [] s t =
      error (shows ''no step from'' \<circ> shows s \<circ> shows '' and '' \<circ> shows t \<circ> shows '' found '')"
  | "check_step c (e # E) s t =
      choice [check_step_rule c e s t, check_step c E s t] <+? shows_sep id shows_nl"

declare check_step.simps[code]

lemma check_step:
  assumes ok: "isOK (check_step chk es s t)"
  shows "\<exists> C \<sigma> p l r. (l,r) \<in> set es \<and> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle> \<and> isOK (chk (l \<cdot> \<sigma>) (r \<cdot> \<sigma>))"
  using assms
proof (induct es)
  case (Cons e es)
  then obtain l r where e:"e = (l,r)" by force
  note mono = ordstep_mono[OF _ subset_refl, of _ "set (e # es)"]
  from Cons show ?case proof(cases "isOK(check_step_rule chk (l,r) s t)")
    case True
    with e check_step_rule [OF this] mono [of "set [e]"] show ?thesis by auto
  next
    case False
    with Cons(2)[unfolded check_step.simps] e have "isOK (check_step chk es s t)" by auto
    with Cons(1) mono show ?thesis by auto
  qed
qed simp

abbreviation "check_rstep \<equiv> check_step (\<lambda> _ _. succeed)"

lemma check_rstep [dest!]:
  assumes "isOK (check_rstep R s t)" 
  shows "(s, t) \<in> rstep (set R)"
  using rstepI and check_step [OF assms] by fast

definition "find_variant_in_trs r R =
  firstM (check_variants_rule r) R <+? (shows_sep id shows_nl)"

lemma find_variant_in_trs_return [elim!]:
  assumes "find_variant_in_trs r R = return r'"
  obtains p where "r' \<in> set R" and "r' = p \<bullet> r"
  using assms by (auto simp: find_variant_in_trs_def dest!: firstM_return)

lemma check_first_variant_trs [elim!]:
  assumes "isOK (find_variant_in_trs r R)"
  obtains r' where "find_variant_in_trs r R = return r'"
  using assms by auto

locale oc_ops =
  fixes check_ord :: "('a::show, string) term \<Rightarrow> ('a, string) term \<Rightarrow> shows check"
begin

abbreviation "check_ordstep \<equiv> check_step check_ord"

abbreviation "check_mem_eq eq E \<equiv>
  check (eq \<in> set E) (shows eq \<circ> shows '' is not an available equation'')"

abbreviation "check_mem_rl \<rho> R \<equiv>
  check (\<rho> \<in> set R) ( shows_rule \<rho> \<circ> shows '' is not an available rule'')"

abbreviation "check_oriented \<equiv> check_allm (\<lambda>(l, r). check_ord l r)"

definition mytest where "mytest =  error (shows ''rule does not match'')"

fun check_step
  where
    "check_step (E, R) (OC_Deduce s t u) = do {
      let S = R @ sym_list E;
      check_rstep S s t <+? (\<lambda>str. shows '' no step from '' \<circ> shows_term s \<circ> shows '' to '' \<circ> shows_term t \<circ> str);
      check_rstep S s u <+? (\<lambda>str. shows '' no step from '' \<circ> shows_term s \<circ> shows '' to '' \<circ> shows_term u \<circ> str);
      return ((t, u) # E, R)
    } <+? (\<lambda>str. shows ''error in deduce step '' \<circ> shows_term t \<circ> shows '' <- '' \<circ> shows_term s \<circ> shows '' -> '' \<circ> shows_term u \<circ> str)"
  | "check_step (E, R) (OC_Orientl s t) = do {
      check_ord s t;
      st \<leftarrow> find_variant_in_trs (s, t) E;
      return (removeAll st E, (s, t) # R)
    } <+? (\<lambda>str. shows ''error in orientl step for '' \<circ> shows_term s \<circ> shows '' -> '' \<circ> shows_term t \<circ> str)"
  | "check_step (E, R) (OC_Orientr s t) = do {
      check_ord t s;
      st \<leftarrow> find_variant_in_trs (s, t) E;
      return (removeAll st E, (t, s) # R)
    } <+? (\<lambda>str. shows ''error in orientr step for '' \<circ> shows_term s \<circ> shows '' -> '' \<circ> shows_term t \<circ> str)"
  | "check_step (E, R) (OC_Delete s) = do {
      ss \<leftarrow> find_variant_in_trs (s, s) E;
      return (removeAll ss E, R)
    } <+? (\<lambda>str. shows ''error in delete step for '' \<circ> shows_term s \<circ> shows '' = '' \<circ> shows_term s \<circ> str)"
  | "check_step (E, R) (OC_Compose s t u) = do {
      st \<leftarrow> find_variant_in_trs (s, t) R;
      let R'' = removeAll st R;
      check_ordstep (R'' @ (sym_list E)) t u  <+? (\<lambda>str. shows '' no ordstep from '' \<circ> shows_term t \<circ> shows '' to '' \<circ> shows_term u \<circ> str);
      return (E, (s, u) # R'')
    } <+? (\<lambda>str. shows ''error in compose step from '' \<circ> shows_term s \<circ> shows '' -> '' \<circ> shows_term t \<circ> shows '' to '' \<circ> 
                                                         shows_term s \<circ> shows '' -> '' \<circ> shows_term u \<circ> str)"
  | "check_step (E, R) (OC_Simplifyl s t u) = do {
      st \<leftarrow> find_variant_in_trs (s, t) E;
      let E'' = removeAll st E;
      check_ordstep (R @ (sym_list E'')) s u  <+? (\<lambda>str. shows '' no ordstep from '' \<circ> shows_term s \<circ> shows '' to '' \<circ> shows_term u \<circ> str);
      return ((u, t) # E'', R)
    } <+? (\<lambda>str. shows ''error in simplifyl step from '' \<circ> shows_term s \<circ> shows '' = '' \<circ> shows_term t \<circ> shows '' to '' \<circ> 
                                                           shows_term u \<circ> shows '' = '' \<circ> shows_term t \<circ> str)"
  | "check_step (E, R) (OC_Simplifyr s t u) = do {
      st \<leftarrow> find_variant_in_trs (s, t) E;
      let E'' = removeAll st E;
      check_ordstep (R @ sym_list E'') t u  <+? (\<lambda>str. shows '' no ordstep from '' \<circ> shows_term t \<circ> shows '' to '' \<circ> shows_term u \<circ> str);
      return ((s, u) # E'', R)
    } <+? (\<lambda>str. shows ''error in simplifyr step from '' \<circ> shows_term s \<circ> shows '' = '' \<circ> shows_term t \<circ> shows '' to '' \<circ>
                                                           shows_term s \<circ> shows '' = '' \<circ> shows_term u \<circ> str)"
  | "check_step (E, R) (OC_Collapse s t u) = do {
      ts \<leftarrow> find_variant_in_trs (t, s) R;
      let R'' = removeAll ts R;
      check_ordstep (R'' @ (sym_list E)) t u  <+? (\<lambda>str. shows '' no ordstep from '' \<circ> shows_term t \<circ> shows '' to '' \<circ> shows_term u \<circ> str);
      return ((u, s) # E, R'')
    } <+? (\<lambda>str. shows ''error in collapse step from '' \<circ> shows_term s \<circ> shows '' -> '' \<circ> shows_term t \<circ> shows '' to '' \<circ> 
                                                          shows_term u \<circ> shows '' = '' \<circ> shows_term t \<circ> str)"

fun check_oc
  where
    "check_oc (E, R) (E', R') [] =
     check_same_set E E' <+? (\<lambda>_. shows '' equation sets do not match'') \<then>
     check_same_set R R' <+? (\<lambda>_. shows '' rule sets do not match'')"
  | "check_oc (E, R) (E', R') (x # xs) = do {
      (E'', R'') \<leftarrow> check_step (E, R) x;
      check_oc (E'', R'') (E', R') xs
    }"

end
lemmas [code] = oc_ops.check_step.simps
lemmas [code] = oc_ops.check_oc.simps
    
locale oc_spec =
  ordered_completion_inf less +
  oc_ops check_ord
  for less :: "('a::show, string) term \<Rightarrow> ('a, string) term \<Rightarrow> bool" (infix "\<succ>" 50)
    and check_ord :: "('a, string) term \<Rightarrow> ('a, string) term \<Rightarrow> shows check" +
  assumes check_ord [simp]: "isOK (check_ord s t) \<longleftrightarrow> s \<succ> t"
begin

lemma check_ordstep:
  assumes "isOK (check_ordstep R s t)"
  shows "(s, t) \<in> ordstep {\<succ>} (set R)"
  using ordstep.intros [of _ _ "set R"] and check_step [OF assms] by auto

lemma check_ordstep_return [dest!]:
  assumes "check_ordstep R s t = return u"
  shows "(s, t) \<in> ordstep {\<succ>} (set R)"
  using assms and check_ordstep by auto
    
lemma check_oriented:
  "isOK (check_oriented rs) \<Longrightarrow> (set rs) \<subseteq> {\<succ>}"
  by auto

lemma check_step_return:
  assumes "check_step (E, R) step = return (E', R')" and "set R \<subseteq> {\<succ>}"
  shows "oKB' (set E, set R) (set E', set R')"
  using assms
proof-
  let ?R = "set R" and ?E = "set E"
  {fix s t RR EE \<pi>
    assume a:"(s, t) \<in> ordstep {\<succ>} (RR \<union> EE\<^sup>\<leftrightarrow>)" "RR \<subseteq> ?R"
    from subst.closedI subst have "subst.closed {\<succ>}" by auto
    from ordstep_subst[OF this] a(1) have "(\<pi> \<bullet> s, \<pi> \<bullet> t) \<in> ordstep {\<succ>} (RR \<union> EE\<^sup>\<leftrightarrow>)"
      unfolding term_apply_subst_Var_Rep_perm[symmetric] by blast
    with ostep_iff_ordstep assms(2) a(2) have "(\<pi> \<bullet> s, \<pi> \<bullet> t) \<in> ostep EE RR" by blast
  }
  note ostep = this
  show ?thesis proof(cases step)
    case (OC_Deduce s t u)
    note ok = assms[unfolded this check_step.simps Let_def]
    then have ids:"R' = R" "E' = (t, u) # E" by auto
    from ok have "isOK(check_rstep (R @ sym_list E) s t) \<and> isOK(check_rstep (R @ sym_list E) s u)" by (simp, force)
    with check_rstep have rsteps:"(s, t) \<in> rstep (?R \<union> ?E\<^sup>\<leftrightarrow>)" "(s, u) \<in> rstep (?R \<union> ?E\<^sup>\<leftrightarrow>)" by auto
    from oKB'.deduce[OF this, of 0] show ?thesis unfolding ids by auto
  next
    case (OC_Orientl s t)
    then show ?thesis
      using assms
      apply (auto simp: Let_def rule_pt.permute_prod.simps elim!: bindE)
    proof-
      fix \<pi>
      assume a:"step = OC_Orientl s t" "s \<succ> t" "(\<pi> \<bullet> s,\<pi> \<bullet> t) \<in> ?E"
      with a(2) have "\<pi> \<bullet> s \<succ> \<pi> \<bullet> t" using less_set_permute by blast
      from oKB'.orientl[OF this a(3), of _ "-\<pi>"]
      show "oKB' (?E, ?R) (?E - {(\<pi> \<bullet> s,\<pi> \<bullet> t)}, insert (s,t) ?R)" by auto
    qed
  next
    case (OC_Orientr s t)
    then show ?thesis
      using assms
      apply (auto simp: Let_def rule_pt.permute_prod.simps elim!: bindE)
    proof-
      fix \<pi>
      assume a:"step = OC_Orientr s t" "t \<succ> s" "(\<pi> \<bullet> s,\<pi> \<bullet> t) \<in> ?E"
      with a(2) have "\<pi> \<bullet> t \<succ> \<pi> \<bullet> s" using less_set_permute by blast
      from oKB'.orientr[OF this a(3), of _ "-\<pi>"]
      show "oKB' (?E, ?R) (?E - {(\<pi> \<bullet> s,\<pi> \<bullet> t)}, insert (t,s) ?R)" by auto
    qed
  next
    case (OC_Delete s)
    then show ?thesis using assms oKB'.delete by (auto simp: rule_pt.permute_prod.simps)
  next
    case (OC_Compose s t u)
    then show ?thesis using assms
      apply (auto simp: Let_def rule_pt.permute_prod.simps elim!: bindE)
    proof-
      fix \<pi>
      let ?Rm = "?R - {(\<pi> \<bullet> s, \<pi> \<bullet> t)}"
      assume a:"(t, u) \<in> ordstep {\<succ>} (?Rm \<union> (set E')\<^sup>\<leftrightarrow>)" "(\<pi> \<bullet> s,\<pi> \<bullet> t) \<in> ?R"
      from ostep[OF a(1)] have "(\<pi> \<bullet> t, \<pi> \<bullet> u) \<in> ostep (set E') ?Rm" by blast
      from oKB'.compose[OF this a(2), of "-\<pi>"]
      show "oKB' (set E', ?R) (set E', insert (s, u) ?Rm)" by force
    qed
  next
    case (OC_Simplifyl s t u)
    then show ?thesis using assms
      apply (auto simp: Let_def rule_pt.permute_prod.simps elim!: bindE)
    proof-
      fix \<pi>
      let ?Em = "?E - {(\<pi> \<bullet> s, \<pi> \<bullet> t)}"
      assume a:"(s, u) \<in> ordstep {\<succ>} (set R' \<union> ?Em\<^sup>\<leftrightarrow>)" "(\<pi> \<bullet> s,\<pi> \<bullet> t) \<in> ?E" "R = R'"
      from ostep[OF a(1)] a(3) have "(\<pi> \<bullet> s, \<pi> \<bullet> u) \<in> ostep ?Em ?R" by blast
      from oKB'.simplifyl[OF this a(2), of "-\<pi>"] a(3)
      show "oKB' (set E, set R') (insert (u, t) ?Em, set R')" by force
    qed
  next
    case (OC_Simplifyr s t u)
    then show ?thesis using assms
      apply (auto simp: Let_def rule_pt.permute_prod.simps elim!: bindE)
    proof-
      fix \<pi>
      let ?Em = "?E - {(\<pi> \<bullet> s, \<pi> \<bullet> t)}"
      assume a:"(t, u) \<in> ordstep {\<succ>} (set R' \<union> ?Em\<^sup>\<leftrightarrow>)" "(\<pi> \<bullet> s,\<pi> \<bullet> t) \<in> ?E" "R = R'"
      from ostep[OF a(1)] a(3) have "(\<pi> \<bullet> t, \<pi> \<bullet> u) \<in> ostep ?Em ?R" by blast
      from oKB'.simplifyr[OF this a(2), of "-\<pi>"] a(3)
      show "oKB' (set E, set R') (insert (s, u) ?Em, set R')" by force
    qed
  next
    case (OC_Collapse s t u)
    then show ?thesis using assms apply (auto simp: Let_def rule_pt.permute_prod.simps elim!: bindE)
    proof-
      fix \<pi>
      let ?Rm = "?R - {(\<pi> \<bullet> t, \<pi> \<bullet> s)}"
      assume a:"(t, u) \<in> ordstep {\<succ>} (?Rm \<union> ?E\<^sup>\<leftrightarrow>)" "(\<pi> \<bullet> t,\<pi> \<bullet> s) \<in> ?R"
      from ostep[OF a(1)] have "(\<pi> \<bullet> t, \<pi> \<bullet> u) \<in> ostep ?E ?Rm" by blast
      from oKB'.collapse[OF this a(2), of "-\<pi>"]
      show "oKB' (set E, set R) (insert (u, s) ?E, ?Rm)" by force
    qed
  qed
qed


lemma oKB'_less:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sup>\<pi> (E', R')" and "R \<subseteq> {\<succ>}"
  shows "R' \<subseteq> {\<succ>}"
proof -
  have [dest]: "(x, y) \<in> R \<Longrightarrow> x \<succ> y" for x y using \<open>R \<subseteq> {\<succ>}\<close> by auto
  note assms(1)
  moreover
  { fix s t \<pi> assume "s \<succ> t"
    then have "\<pi> \<bullet> s \<succ> \<pi> \<bullet> t" using less_set_permute by auto }
  moreover
  { fix s t u \<pi> assume "(t, u) \<in> rstep (R - {(s, t)})" and "(s, t) \<in> R"
    then have "s \<succ> u" by (induct t u) (blast intro: subst ctxt dest: trans)
    then have "\<pi> \<bullet> s \<succ> \<pi> \<bullet> u" using less_set_permute by auto }
  moreover
  { fix s t u \<pi> assume "(t, u) \<in> ordstep {\<succ>} (E\<^sup>\<leftrightarrow>)" and "(s, t) \<in> R"
    then have "s \<succ> u" by (cases) (auto dest: trans ctxt)
    then have "\<pi> \<bullet> s \<succ> \<pi> \<bullet> u" using less_set_permute by auto }
  ultimately show ?thesis by (cases, (fastforce simp: ostep_def)+)
qed

lemma oKB'_rtrancl_less:
  assumes "oKB'\<^sup>*\<^sup>* (E, R) (E', R')" and "R \<subseteq> {\<succ>}"
  shows "R' \<subseteq> {\<succ>}"
  using assms by (induct "(E, R)" "(E', R')" arbitrary: E' R') (auto dest!: oKB'_less)

lemma oKB_ER_permuted:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sup>\<pi> (E', R')"
  shows "E \<union> R \<subseteq> Id \<union> E' \<union> R' \<union> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow> \<union> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow> O (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>"
  using assms
proof-
  from ostep_imp_rstep_sym have orstep:"\<And>s t.(s, t) \<in> ostep E' R' \<Longrightarrow> (s, t) \<in> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>" by blast
  let ?step2 = "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow> O (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>"
  show ?thesis using assms
  proof (cases)
    case (simplifyl s u t \<pi>)
    then have su:"(s, u) \<in> ostep E' R'"
      by (intro ostep_mono [THEN subsetD, OF _ _ simplifyl(3)]) auto
    from simplifyl have "(\<pi> \<bullet> u, \<pi> \<bullet> t) \<in> E'" by auto
    from rstep.rule[OF this] have "(u,t) \<in> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>" using rstep_permute_iff by blast
    with orstep[OF su] have "(s, t) \<in> ?step2" by blast
    then show ?thesis unfolding simplifyl by blast
  next
    case (simplifyr t u s \<pi>)
    then have tu:"(t, u) \<in> ostep E' R'"
      by (intro ostep_mono [THEN subsetD, OF _ _ simplifyr(3)]) auto
    from simplifyr have "(\<pi> \<bullet> s, \<pi> \<bullet> u) \<in> E'" by auto
    from rstep.rule[OF this] have "(s, u) \<in> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>" using rstep_permute_iff by blast
    with orstep[OF tu] have "(s, t) \<in> ?step2" unfolding simplifyr by blast
    then show ?thesis unfolding simplifyr by blast
  next
    case (orientl s t \<pi>)
    have "(s,t) \<in> (rstep R')" unfolding orientl using rstep_permute_iff
      by (metis UnI2 rstep.rule rstep_union singletonI)
    then show ?thesis using orientl by fast
  next
    case (orientr s t \<pi>)
    have "(s,t) \<in> (rstep R')" unfolding orientr using rstep_permute_iff
      by (metis UnI2 rstep.rule rstep_union singletonI)
    then show ?thesis using orientr by fast
  next
    case (compose t u s \<pi>)
    then have tu:"(t, u) \<in> ostep E' R'"
      by (intro ostep_mono [THEN subsetD, OF _ _ compose(3)]) auto
    from compose have "(\<pi> \<bullet> s, \<pi> \<bullet> u) \<in> R'" by auto
    from rstep.rule[OF this] have "(s, u) \<in> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>" using rstep_permute_iff by blast
    with orstep[OF tu] have "(s, t) \<in> ?step2" by blast
    then show ?thesis unfolding compose by blast
  next
    case (collapse t u s \<pi>)
    then have tu:"(t, u) \<in> ostep E' R'"
      by (intro ostep_mono [THEN subsetD, OF _ _ collapse(3)]) auto
    from collapse have "(\<pi> \<bullet> u, \<pi> \<bullet> s) \<in> E'" by auto
    from rstep.rule[OF this] have "(u,s) \<in> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>" using rstep_permute_iff by blast
    with orstep[OF tu] have "(t, s) \<in> ?step2" by auto
    then show ?thesis unfolding collapse by blast
  qed auto
qed

lemma oKB_ER_permuted_rev:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sup>\<pi> (E', R')"
  shows "E' \<union> R' \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow> \<union> (rstep (E \<union> R))\<^sup>\<leftrightarrow> O (rstep (E \<union> R))\<^sup>\<leftrightarrow>"
proof-
  from ostep_imp_rstep_sym have orstep:"\<And>s t.(s, t) \<in> ostep E R \<Longrightarrow> (s, t) \<in> (rstep (E \<union> R))\<^sup>\<leftrightarrow>" by blast
  let ?step2 = "(rstep (E \<union> R))\<^sup>\<leftrightarrow> O (rstep (E \<union> R))\<^sup>\<leftrightarrow>"
show ?thesis using assms
proof (cases)
  case (simplifyl s u t \<pi>)
  have "(s, u) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ simplifyl(3)]) auto
  with simplifyl(4) orstep have "(u, t) \<in> ?step2" by fast
  then show ?thesis unfolding simplifyl by auto
next
  case (simplifyr t u s \<pi>)
  have "(t, u) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ simplifyr(3)]) auto
  with simplifyr(4) orstep have "(s, u) \<in> ?step2" by fast
  then show ?thesis unfolding simplifyr by auto
next
  case (deduce u s t \<pi>)
  then have "(u, s) \<in> (rstep (E \<union> R))\<^sup>\<leftrightarrow>" "(u, t) \<in> (rstep (E \<union> R))\<^sup>\<leftrightarrow>" by auto
  then have "(s, t) \<in> ?step2" by blast
  with deduce show ?thesis by auto
next
  case (collapse t u s \<pi>)
  have "(t, u) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ collapse(3)]) auto
  with collapse(4) orstep have "(u, s) \<in> ?step2" by fast
  then show ?thesis unfolding collapse by auto
next
  case (compose t u s \<pi>)
  have "(t, u) \<in> ostep E R"
    by (intro ostep_mono [THEN subsetD, OF _ _ compose(3)]) auto
  with compose(4) orstep have "(s, u) \<in> ?step2" by fast
  then show ?thesis unfolding compose by auto
qed auto
qed

lemma oKB_permuted_rstep_subset:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sup>\<pi> (E', R')"
  shows "rstep (E \<union> R) \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
proof-
  let ?C = "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  have 1:"rstep (Id \<union> E' \<union> R') \<subseteq> ?C" by fast
  have 2:"rstep ((rstep (E' \<union> R'))\<^sup>\<leftrightarrow>) \<subseteq> ?C" by fast
  from relcomp_mono[OF 2 2] have 3:"(rstep (E' \<union> R'))\<^sup>\<leftrightarrow> O (rstep (E' \<union> R'))\<^sup>\<leftrightarrow> \<subseteq> ?C"
    unfolding conversion_O_conversion by auto
  from rstep_mono[OF 3] rstep_rstep have 3:"rstep ((rstep (E' \<union> R'))\<^sup>\<leftrightarrow> O (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>) \<subseteq> ?C"
    using conversion_O_conversion
    by (metis relcomp_mono[OF 2 2] conversion_O_conversion rstep_relcomp_idemp1 rstep_simps(5))
  show ?thesis using oKB_ER_permuted [OF oKB, THEN rstep_mono]
  unfolding rstep_union[of _ "_ O _"] rstep_union[of _ "( _)\<^sup>\<leftrightarrow>"] 
  using 1 2 3 by blast
qed

lemma oKB_conversion_permuted_subset:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sup>\<pi> (E', R')"
  shows "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*"
  using conversion_mono [OF oKB_permuted_rstep_subset [OF assms]] by auto

lemma oKB_permuted_rstep_subset_rev:
  assumes oKB: "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sup>\<pi> (E', R')"
  shows "rstep (E' \<union> R') \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
proof-
  let ?C = "(rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  have 1:"rstep ((rstep (E \<union> R))\<^sup>\<leftrightarrow>) \<subseteq> ?C" by fast
  from relcomp_mono[OF 1 1] have 2:"(rstep (E \<union> R))\<^sup>\<leftrightarrow> O (rstep (E \<union> R))\<^sup>\<leftrightarrow> \<subseteq> ?C"
    unfolding conversion_O_conversion by auto
  from rstep_mono[OF 2] rstep_rstep have 3:"rstep ((rstep (E \<union> R))\<^sup>\<leftrightarrow> O (rstep (E \<union> R))\<^sup>\<leftrightarrow>) \<subseteq> ?C"
    using conversion_O_conversion
    by (metis relcomp_mono[OF 1 1] conversion_O_conversion rstep_relcomp_idemp1 rstep_simps(5))
  from oKB_ER_permuted_rev[OF assms, THEN rstep_mono] show ?thesis
    unfolding rstep_union[of _ "_ O _"] using 1 3 by blast
qed

lemma oKB_conversion_permuted_subset_rev:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sup>\<pi> (E', R')"
  shows "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using conversion_mono [OF oKB_permuted_rstep_subset_rev [OF assms]] by auto

lemma oKB_step_conversion_permuted:
  assumes "(E, R) \<turnstile>\<^sub>o\<^sub>K\<^sub>B\<^sup>\<pi> (E', R')"
  shows "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using oKB_conversion_permuted_subset[OF assms] oKB_conversion_permuted_subset_rev[OF assms] by auto

lemma oKB_steps_conversion_permuted:
  assumes "oKB'\<^sup>*\<^sup>* (E, R) (E', R')"
  shows "(rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
  using assms
  by (induct "(E, R)" "(E', R')" arbitrary: E' R')
    (force dest!: oKB_step_conversion_permuted)+

lemma oKB_steps_conversion_FGROUND:
  assumes "oKB'\<^sup>*\<^sup>* (E, R) (E', R')"
  shows "FGROUND F ((rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*) = FGROUND F ((rstep (E' \<union> R'))\<^sup>\<leftrightarrow>\<^sup>*)"
  unfolding FGROUND_def using oKB_steps_conversion_permuted[OF assms] by auto

lemma E_ord_ordstep:"rstep (E_ord (\<succ>) E) = ordstep {\<succ>} (E\<^sup>\<leftrightarrow>)"
proof(rule, rule)
  fix s t
  assume "(s,t) \<in> rstep (E_ord (\<succ>) E)"
  from rstepE[OF this] obtain l r C \<tau> \<sigma> where lr:"(l,r) \<in> E\<^sup>\<leftrightarrow>" and ts:"l\<cdot>\<tau> \<succ> r\<cdot>\<tau>" "s = C\<langle>l\<cdot>\<tau>\<cdot>\<sigma>\<rangle>" "t = C\<langle>r\<cdot>\<tau>\<cdot>\<sigma>\<rangle>"
    unfolding E_ord_def mem_Collect_eq
    by (metis Pair_inject)
  from ts subst[OF ts(1)] have ts:"l\<cdot>(\<tau> \<circ>\<^sub>s \<sigma>) \<succ> r\<cdot>(\<tau> \<circ>\<^sub>s \<sigma>)" "s = C\<langle>l\<cdot>(\<tau> \<circ>\<^sub>s \<sigma>)\<rangle>" "t = C\<langle>r\<cdot>(\<tau> \<circ>\<^sub>s \<sigma>)\<rangle>"
    unfolding subst_subst_compose by auto
  show "(s,t) \<in> ordstep {\<succ>} (E\<^sup>\<leftrightarrow>)" unfolding ordstep.simps
    by(rule exI[of _ l],rule exI, rule exI,rule exI, rule exI[of _ "\<tau> \<circ>\<^sub>s \<sigma>"], insert lr ts, auto)
next
  show "ordstep {\<succ>} (E\<^sup>\<leftrightarrow>) \<subseteq> rstep (E_ord (\<succ>) E)"
  proof
    fix s t
    assume "(s, t) \<in> ordstep {\<succ>} (E\<^sup>\<leftrightarrow>)"
    then obtain l r C \<tau> where lr:"(l,r) \<in> E\<^sup>\<leftrightarrow>" and ts:"l\<cdot>\<tau> \<succ> r\<cdot>\<tau>" "s = C\<langle>l\<cdot>\<tau>\<rangle>" "t = C\<langle>r\<cdot>\<tau>\<rangle>" unfolding ordstep.simps by auto
    then have "(l\<cdot>\<tau>,r\<cdot>\<tau>) \<in> E_ord (\<succ>) E" unfolding E_ord_def by auto
    with lr show "(s, t) \<in> rstep (E_ord (\<succ>) E)" unfolding ts by auto
  qed
qed

lemma FGROUND_rstep_FGROUND_ostep:
  assumes "R \<subseteq> {\<succ>}" and gtotal: "\<And>s t. ground s \<Longrightarrow> ground t \<Longrightarrow> s = t \<or> s \<succ> t \<or> t \<succ> s"
  shows "FGROUND F ((rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*) = FGROUND F ((ostep E R)\<^sup>\<leftrightarrow>\<^sup>*)"
proof(rule, rule)
  fix s t
  assume "(s,t) \<in> FGROUND F ((rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*)"
  then have fs:"funas_term s \<subseteq> F" "funas_term t \<subseteq> F" and g:"ground s" "ground t" "(s,t) \<in> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding FGROUND_def fground_def by auto
  note gconv = gterm_conv_GROUND_conv[OF g]
  interpret gtotal_reduction_order by (insert gtotal, standard, auto)
  from GROUND_rstep_ordstep[of R E] have "GROUND (rstep (R \<union> E\<^sup>\<leftrightarrow>)) \<subseteq> (GROUND (ostep E R))\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding ostep_def E_ord_ordstep[symmetric] rstep_union by auto
  then have "GROUND (rstep (E \<union> R)) \<subseteq> (GROUND (ostep E R))\<^sup>\<leftrightarrow>\<^sup>*" unfolding rstep_union GROUND_def by auto
  from conversion_mono[OF this] have "(GROUND (rstep (E \<union> R)))\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (GROUND (ostep E R))\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding conversion_conversion_idemp by auto
  with gconv have "(s,t) \<in> (GROUND (ostep E R))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  with conversion_mono[OF GROUND_subset] have "(s,t) \<in> (ostep E R)\<^sup>\<leftrightarrow>\<^sup>*" by auto
  with fs g show "(s,t) \<in> FGROUND F ((ostep E R)\<^sup>\<leftrightarrow>\<^sup>*)" unfolding FGROUND_def fground_def by auto
next
  note ostep_conv = ostep_iff_ordstep[OF assms(1), symmetric]
  from ordstep_imp_rstep[of _ _ "{\<succ>}" "R \<union> _\<^sup>\<leftrightarrow>", unfolded ostep_conv] have 1:"ostep E R \<subseteq> rstep (R \<union> E\<^sup>\<leftrightarrow>)" by auto
  have "rstep (R \<union> E\<^sup>\<leftrightarrow>) \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  with 1 have "ostep E R \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*" by auto
  from conversion_mono[OF this] have "(ostep E R)\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> (rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*"
    unfolding conversion_conversion_idemp by auto
  then show "FGROUND F ((ostep E R)\<^sup>\<leftrightarrow>\<^sup>*) \<subseteq> FGROUND F ((rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*)" unfolding FGROUND_def by auto
qed


lemma oKB'_rtrancl_FGROUND_conversion:
  assumes "oKB'\<^sup>*\<^sup>* (E, R) (E', R')" and
    R:"R \<subseteq> {\<succ>}" and
    gtotal: "\<And>s t. ground s \<Longrightarrow> ground t \<Longrightarrow> s = t \<or> s \<succ> t \<or> t \<succ> s"
  shows "FGROUND F ((rstep (E \<union> R))\<^sup>\<leftrightarrow>\<^sup>*) = FGROUND F ((ostep E' R')\<^sup>\<leftrightarrow>\<^sup>*)"
proof-
  from assms oKB'_rtrancl_less have "R' \<subseteq> {\<succ>}" by auto
  from oKB_steps_conversion_FGROUND[OF assms(1)] FGROUND_rstep_FGROUND_ostep[OF this] gtotal show ?thesis by auto
qed

lemma check_oc:
  assumes ok: "isOK (check_oc (E, R) (E', R') xs)" and R:"isOK (check_oriented R)"
  shows "oKB'\<^sup>*\<^sup>* (set E, set R) (set E', set R')"
proof-
  from check_oriented[OF R] have R: "set R \<subseteq> {\<succ>}" by auto
  with ok show ?thesis
  proof (induct xs arbitrary: E R)
    case Nil
    then show ?case by auto
  next
    case (Cons x xs)
    then obtain E'' and R'' where 1: "check_step (E, R) x = return (E'', R'')"
      and 2: "isOK (check_oc (E'', R'') (E', R') xs)" by auto
    note step = check_step_return[OF 1 Cons(3)]
    from step[THEN oKB'_less] Cons(3) have less:"set R'' \<subseteq> {\<succ>}" by auto
    from Cons.hyps [OF 2 less] step show ?case by auto
  qed
qed

end


locale oc_spec_extension = 
oc:oc_spec less1 check_ord1 +
oc':oc_spec less2 check_ord2
for less1 less2 :: "('a::show, string) term \<Rightarrow> ('a, string) term \<Rightarrow> bool" and
    check_ord1 check_ord2  check_ord :: "('a, string) term \<Rightarrow> ('a, string) term \<Rightarrow> shows check" +
assumes less_extension:"\<And>s t. less1 s t \<Longrightarrow> less2 s t"
begin

lemma ostep:"(s,t) \<in> oc.ostep E R \<Longrightarrow> (s,t) \<in> oc'.ostep E R"
  unfolding oc.ostep_def oc'.ostep_def Un_iff ordstep.simps
  using less_extension by fast

lemma oKB'_step:
  assumes "oc.oKB' (E,R) (E',R')"
  shows "oc'.oKB' (E,R) (E',R')"
  using assms
proof(cases)
  case (deduce s t u p)
  show ?thesis unfolding deduce
    using deduce(3) deduce(4) oc'.oKB'.deduce by blast
next
  case (orientl s t p)
  show ?thesis unfolding orientl
    using less_extension[OF orientl(3)] orientl(4) oc'.oKB'.orientl by blast
next
  case (orientr t s p)
  show ?thesis unfolding orientr
    using less_extension[OF orientr(3)] orientr(4) oc'.oKB'.orientr by blast
next
  case (delete s)
  show ?thesis unfolding delete using delete(3) oc'.oKB'.delete by blast
next
  case (compose t u s p)
  show ?thesis unfolding compose
    using ostep[OF compose(3)] compose(4) oc'.oKB'.compose by blast
next
  case (simplifyl s u t p)
  show ?thesis unfolding simplifyl
    using ostep[OF simplifyl(3)] simplifyl(4) oc'.oKB'.simplifyl by blast
next
  case (simplifyr t u s p)
  show ?thesis unfolding simplifyr
    using ostep[OF simplifyr(3)] simplifyr(4) oc'.oKB'.simplifyr by blast
next
  case (collapse t u s p)
  show ?thesis unfolding collapse
    using ostep[OF collapse(3)] collapse(4) oc'.oKB'.collapse by blast
qed

lemma oKB'_steps:
  assumes "oc.oKB'\<^sup>*\<^sup>* (E,R) (E',R')"
  shows "oc'.oKB'\<^sup>*\<^sup>* (E,R) (E',R')"
  using assms
  by (induct "(E, R)" "(E', R')" arbitrary: E' R')
     (force dest!: oKB'_step)+
end

locale gcr_ops =
  fixes xvar yvar :: "'v::{show, infinite} \<Rightarrow> 'v"
    and check_joinable :: "('f::show, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
    and check_instance :: "('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
begin

definition "check_ooverlap E R \<rho>\<^sub>1 \<rho>\<^sub>2 p =
  (case mgu_var_disjoint_generic xvar yvar (fst \<rho>\<^sub>1) (fst \<rho>\<^sub>2 |_ p) of
    None \<Rightarrow> succeed
  | Some (\<sigma>\<^sub>1, \<sigma>\<^sub>2) \<Rightarrow>
    let s = replace_at (fst \<rho>\<^sub>2 \<cdot> \<sigma>\<^sub>2) p (snd \<rho>\<^sub>1 \<cdot> \<sigma>\<^sub>1)  in
    let t = snd \<rho>\<^sub>2 \<cdot> \<sigma>\<^sub>2 in
    choice [
      check_instance E R s t <+? (\<lambda>e. shows ''CP '' \<circ> shows_term s \<circ> shows '' = '' \<circ> shows_term t \<circ> shows '' not an instance of an equation'' \<circ> e),
      check_joinable E R s t <+? (\<lambda>e. shows ''CP '' \<circ> shows_term s \<circ> shows '' = '' \<circ> shows_term t \<circ> shows '' is not joinable'' \<circ> e)
    ] <+? shows_sep id shows_nl)"

text \<open>Check extended critical pairs.\<close>
definition
  "check_ECPs E R =
    (let E' = sym_list E; S = List.union R E' in
    check_allm (\<lambda>\<rho>\<^sub>2. let l\<^sub>2 = fst \<rho>\<^sub>2 in check_allm (\<lambda>\<rho>\<^sub>1. check_allm (\<lambda>p.
      check_ooverlap E' R \<rho>\<^sub>1 \<rho>\<^sub>2 p) (funposs_list l\<^sub>2)) S) S)"

end
  
lemmas [code] =
  gcr_ops.check_ooverlap_def gcr_ops.check_ECPs_def
 
definition "Rules E R = List.union R (sym_list E)"

lemma set_Rules [simp]: "set (Rules E R) = set R \<union> (set E)\<^sup>\<leftrightarrow>" by (auto simp: Rules_def)

locale gcr_spec =
  gcr_ops xvar yvar check_joinable check_instance +
  gtotal_reduction_order_inf less
  for xvar :: "'v::{show, infinite} \<Rightarrow> 'v"
  and yvar :: "'v \<Rightarrow> 'v"
  and check_joinable :: "('f::show, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
  and check_instance :: "('f, 'v) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> shows check"
  and less :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" (infix "\<succ>" 50)
  and F :: "('f \<times> nat) list" +
  assumes joinable: "set R \<subseteq> {\<succ>} \<Longrightarrow> isOK (check_joinable E R s t) \<Longrightarrow>
      (s, t) \<in> (ordstep {\<succ>} (set (Rules E R)))\<^sup>\<down>"
    and "instance": "isOK (check_instance E R s t) \<Longrightarrow> \<exists>l r \<sigma>. (l, r) \<in> set (Rules E R) \<and> s = l \<cdot> \<sigma> \<and> t = r \<cdot> \<sigma>"
    and ren: "inj xvar" "inj yvar" "range xvar \<inter> range yvar = {}"
begin

lemma check_ECPs_ooverlap:
  assumes "isOK (check_ECPs E R)"
    and "ooverlap {\<succ>} (set (Rules E R)) r r' p \<mu> s t"
    and R: "set R \<subseteq> {\<succ>}"
  shows "(s, t) \<in> (ordstep {\<succ>} (set (Rules E R)))\<^sup>\<down> \<or>
    (\<exists>l r \<sigma>. (l, r) \<in> (set (Rules E R))\<^sup>\<leftrightarrow> \<and> s = l \<cdot> \<sigma> \<and> t = r \<cdot> \<sigma>)"
proof -
  let ?R = "set (Rules E R)"
  from assms(2) obtain \<pi>\<^sub>1 and \<pi>\<^sub>2 where rules': "\<pi>\<^sub>1 \<bullet> r \<in> ?R" "\<pi>\<^sub>2 \<bullet> r' \<in> ?R"
    and disj: "vars_rule r \<inter> vars_rule r' = {}"
    and p': "p \<in> funposs (fst r')"
    and mgu': "mgu (fst r) (fst r' |_ p) = Some \<mu>"
    and "\<not> (snd r \<cdot> \<mu> \<succ> fst r \<cdot> \<mu>)" "\<not> (snd r' \<cdot> \<mu> \<succ> fst r' \<cdot> \<mu>)"
    and t: "t = snd r' \<cdot> \<mu>"
    and s: "s = replace_at (fst r' \<cdot> \<mu>) p (snd r \<cdot> \<mu>)"
    unfolding ooverlap_def by fast
  define \<rho>\<^sub>1 and \<rho>\<^sub>2 where "\<rho>\<^sub>1 = \<pi>\<^sub>1 \<bullet> r" and "\<rho>\<^sub>2 = \<pi>\<^sub>2 \<bullet> r'"
  have p: "p \<in> funposs (fst \<rho>\<^sub>2)"
    and "\<rho>\<^sub>1 \<in> ?R" and "\<rho>\<^sub>2 \<in> ?R"
    using p' and rules' by (auto simp: \<rho>\<^sub>1_def \<rho>\<^sub>2_def rule_pt.fst_eqvt [symmetric])
  with assms have *: "isOK (check_ooverlap (sym_list E) R \<rho>\<^sub>1 \<rho>\<^sub>2 p)" by (auto simp: check_ECPs_def)
      
  have "(fst r' |_ p) \<cdot> \<mu> = fst r \<cdot> \<mu>" using mgu' [THEN mgu_sound] by (auto simp: is_imgu_def)
  then have "(fst r' |_ p) \<cdot> (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>2) \<circ> Rep_perm \<pi>\<^sub>2) =
    fst r \<cdot> (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>1) \<circ> Rep_perm \<pi>\<^sub>1)"
    by (simp add: o_assoc [symmetric] Rep_perm_add [symmetric] Rep_perm_0)
  then have "fst \<rho>\<^sub>2 |_ p \<cdot> (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>2)) = fst \<rho>\<^sub>1 \<cdot> (\<mu> \<circ> Rep_perm (-\<pi>\<^sub>1))"
    using p [THEN funposs_imp_poss]
    by (auto simp: \<rho>\<^sub>1_def \<rho>\<^sub>2_def permute_term_subst_apply_term [symmetric] eqvt)
  from mgu_var_disjoint_generic_complete [OF ren this[symmetric]] obtain \<mu>\<^sub>1 \<mu>\<^sub>2 \<delta>
    where mgu: "mgu_var_disjoint_generic xvar yvar (fst \<rho>\<^sub>1) (fst \<rho>\<^sub>2 |_ p) = Some (\<mu>\<^sub>1, \<mu>\<^sub>2)"
    and **: "\<mu> \<circ> Rep_perm (- \<pi>\<^sub>1) = \<mu>\<^sub>1 \<circ>\<^sub>s \<delta>" "\<mu> \<circ> Rep_perm (- \<pi>\<^sub>2) = \<mu>\<^sub>2 \<circ>\<^sub>s \<delta>" by blast

  have "vars_term (fst r) \<subseteq> vars_rule r" and "vars_term (fst r' |_ p) \<subseteq> vars_rule r'"
    using p' [THEN funposs_imp_poss, THEN vars_term_subt_at] by (auto simp: vars_defs)
  from mgu_imp_mgu_var_disjoint [OF mgu' ren disj this finite_vars_rule finite_vars_rule, of \<pi>\<^sub>1 \<pi>\<^sub>2]
  obtain \<mu>' and \<pi>
    where left: "\<forall>x\<in>vars_rule r. \<mu> x = (sop \<pi>\<^sub>1 \<circ>\<^sub>s (\<mu>' \<circ> xvar) \<circ>\<^sub>s sop (- \<pi>)) x" (is "\<forall>x\<in>_. \<mu> x = ?\<pi>\<^sub>1 x")
      and right: "\<forall>x\<in>vars_rule r'. \<mu> x = (sop \<pi>\<^sub>2 \<circ>\<^sub>s (\<mu>' \<circ> yvar) \<circ>\<^sub>s sop (- \<pi>)) x" (is "\<forall>x\<in>_. \<mu> x = ?\<pi>\<^sub>2 x")
      and "mgu (\<pi>\<^sub>1 \<bullet> (fst r) \<cdot> (Var \<circ> xvar)) (\<pi>\<^sub>2 \<bullet> (fst r' |_ p) \<cdot> (Var \<circ> yvar)) = Some \<mu>'"
    by blast
  then have "mgu_var_disjoint_generic xvar yvar (fst \<rho>\<^sub>1) (fst \<rho>\<^sub>2 |_ p) = Some (\<mu>' \<circ> xvar, \<mu>' \<circ> yvar)"
    and \<mu>\<^sub>1: "\<mu>\<^sub>1 = \<mu>' \<circ> xvar" and \<mu>\<^sub>2: "\<mu>\<^sub>2 = \<mu>' \<circ> yvar"
    using p' [THEN funposs_imp_poss] and mgu
    unfolding \<rho>\<^sub>1_def \<rho>\<^sub>2_def mgu_var_disjoint_generic_def 
    by (auto simp del: mgu.simps
        simp: map_vars_term_eq mgu_var_disjoint_generic_def \<rho>\<^sub>1_def \<rho>\<^sub>2_def eqvt)

  let ?t = "snd \<rho>\<^sub>2 \<cdot> \<mu>\<^sub>2"
  let ?t' = "snd r' \<cdot> \<mu>"
  let ?s = "replace_at (fst \<rho>\<^sub>2 \<cdot> \<mu>\<^sub>2) p (snd \<rho>\<^sub>1 \<cdot> \<mu>\<^sub>1)"
  let ?s' = "replace_at (fst r' \<cdot> \<mu>) p (snd r \<cdot> \<mu>)"
 
  have "fst r' \<cdot> \<mu> = fst r' \<cdot> ?\<pi>\<^sub>2" and snd_r: "snd r' \<cdot> \<mu> = snd r' \<cdot> ?\<pi>\<^sub>2"
    using right unfolding term_subst_eq_conv by (auto simp: vars_defs)
  then have 1: "fst \<rho>\<^sub>2 \<cdot> \<mu>\<^sub>2 = \<pi> \<bullet> (fst r' \<cdot> \<mu>)"
    and t': "?t' = -\<pi> \<bullet> ?t" by (auto simp: \<mu>\<^sub>2 \<rho>\<^sub>2_def eqvt)
  
  have "snd r \<cdot> \<mu> = snd r \<cdot> ?\<pi>\<^sub>1"
    using left unfolding term_subst_eq_conv by (auto simp: vars_defs)
  then have 2: "snd \<rho>\<^sub>1 \<cdot> \<mu>\<^sub>1 = \<pi> \<bullet> (snd r \<cdot> \<mu>)" by (auto simp: \<mu>\<^sub>1 \<rho>\<^sub>1_def eqvt)
  then have s': "?s' = -\<pi> \<bullet> ?s"
    using p' [THEN funposs_imp_poss]
    unfolding 1 by (auto simp: eqvt)
  from snd_r have 2: "snd \<rho>\<^sub>2 \<cdot> \<mu>\<^sub>2 = \<pi> \<bullet> (snd r' \<cdot> \<mu>)" by (auto simp: \<mu>\<^sub>2 \<rho>\<^sub>2_def eqvt)

  consider (join) "isOK (check_joinable (sym_list E) R ?s ?t)"
    | (inst) "isOK (check_instance (sym_list E) R ?s ?t)"
    using * and mgu unfolding check_ooverlap_def by (auto simp: check_ooverlap_def)
  then show ?thesis
  proof (cases)
    case join
    from joinable [OF R join] have "(?s, ?t) \<in> (ordstep {\<succ>} ?R)\<^sup>\<down>" by simp
    from join_subst [OF subst_closed_ordstep [OF subst_closed_less] this, of "sop (-\<pi>)"]
    have "(s, t) \<in> (ordstep {\<succ>} ?R)\<^sup>\<down>"
      using p [THEN funposs_imp_poss]
      unfolding s t s' using term_pt.permute_flip [OF 1] term_pt.permute_flip [OF 2]
      by (simp add: eqvt [symmetric])
    then show ?thesis by blast
  next
    case inst
    from "instance" [OF inst] obtain u and v and \<sigma>
      where uv: "(u, v) \<in> ?R\<^sup>\<leftrightarrow>" and "?s = u \<cdot> \<sigma>" and "?t = v \<cdot> \<sigma>" by auto
    then have "?s' = u \<cdot> (\<sigma> \<circ>\<^sub>s (sop (- \<pi>)))"
      and "?t' = v \<cdot> (\<sigma> \<circ>\<^sub>s (sop (- \<pi>)))" by (auto simp: s' t')
    with uv show ?thesis unfolding s t by blast
  qed
qed

lemma check_ECPs:
  assumes R: "set R \<subseteq> {\<succ>}" and ok: "isOK (check_ECPs E R)"
  shows "GCR (ordstep {\<succ>} (set (Rules E R)))"
proof -
  { fix s t assume "(s, t) \<in> set (Rules E R)"
    with R have "s \<succ> t \<or> (t, s) \<in> set (Rules E R)" by auto }
  then show ?thesis
    using check_ECPs_ooverlap [OF ok _ R]
    by (intro GCR_ordstep) auto
qed

end

definition "check_joinable R s t =
  (case (compute_rstep_NF R s, compute_rstep_NF R t) of
    (Some u, Some v) \<Rightarrow> check (u = v) (shows ''normal forms differ'')
  | _ \<Rightarrow> error (shows ''error: check_joinable''))"
  
lemma check_joinable:
  assumes "isOK (check_joinable R s t)"
  shows "(s, t) \<in> (rstep (set R))\<^sup>\<down>"
  using assms
  by (auto simp: check_joinable_def split: option.splits dest: compute_rstep_NF_sound)

definition "check_rule_instance r r' =
  (case match_list Var [(fst r', fst r), (snd r', snd r)] of
    Some _ \<Rightarrow> succeed
  | None \<Rightarrow> error (shows ''rules do not match''))"
  
lemma check_rule_instance [simp]:
  "isOK (check_rule_instance (l, r) (l', r')) \<longleftrightarrow> (\<exists>\<sigma>. l = l' \<cdot> \<sigma> \<and> r = r' \<cdot> \<sigma>)"
  by (auto simp: check_rule_instance_def split: option.splits dest!: match_list_sound match_list_complete)

definition "check_instance E s t =
  check_exm (check_rule_instance (s, t)) E (shows_sep id shows_nl)"
  
lemma check_instance [simp]:
   "isOK (check_instance E s t) \<longleftrightarrow> (\<exists>l r \<sigma>. (l, r) \<in> set E \<and> s = l \<cdot> \<sigma> \<and> t = r \<cdot> \<sigma>)"
  by (auto simp: check_instance_def) (insert check_rule_instance, blast+)

context oc_spec
begin

lemma gcr_spec:
  assumes "gtotal_reduction_order (\<succ>)"
  shows "gcr_spec x_var y_var (\<lambda>_. check_joinable) (\<lambda>E _. check_instance E) (\<succ>)"
proof -
  interpret gtotal_reduction_order "(\<succ>)" by fact
  interpret gcr_spec x_var y_var
    "(\<lambda>_. check_joinable)" "(\<lambda>E _. check_instance E)"
    apply (unfold_locales)
        apply (drule check_joinable)
        apply (rule join_mono [THEN subsetD, of "rstep (set R)" "ordstep {\<succ>} (set (Rules E R))" for E R])
         apply (simp add: ordstep_Un ordstep_rstep_conv subst_closed_less)
        apply assumption
       apply force+
    done
  show ?thesis ..
qed

end

definition "check_ECPs =
  gcr_ops.check_ECPs x_var y_var (\<lambda>_. check_joinable) (\<lambda>E _. check_instance E)"

definition "check_FGCR ro F E R = do {
  redord.valid ro;
  check_subseteq (funas_trs_list (List.union E R)) F
    <+? (\<lambda>f. shows ''the function symbol '' \<circ> shows f \<circ> shows '' does not occur in the TRS\<newline>'');
  check_ECPs E R
}"
    
lemma check_FGCR:
  fixes F :: "('f::{show,compare_order} \<times> nat) list"
  assumes "isOK (check_FGCR (create_KBO_redord r F) F E R)"
    and R: "set R \<subseteq> {(s, t). redord.less (create_KBO_redord r F) s t}"
  shows "CR (FGROUND (set F) (ordstep ({(s, t). redord.less (create_KBO_redord r F) s t}) (set (Rules E R))))"
proof -
  let ?F = "set F"
  let ?R = "set (Rules E R)"

  define less :: "('f, string) term \<Rightarrow> ('f, string) term \<Rightarrow> bool" and c :: "'f"
    where
      "less = redord.less (create_KBO_redord r F)" and
      "c = redord.min_const (create_KBO_redord r F)"

  have valid: "isOK (redord.valid (create_KBO_redord r F))"
    and ecps: "isOK (check_ECPs E R)"
    and F: "funas_trs ?R \<subseteq> ?F"
    using assms by (auto simp: check_FGCR_def)

  from KBO_redord.valid [OF valid] obtain less'
    where ro: "reduction_order less"
      and c: "(c, 0) \<in> ?F"
      and fground: "\<forall>s t. fground ?F s \<and> fground ?F t \<longrightarrow> s = t \<or> less s t \<or> less t s"
      and ro': "gtotal_reduction_order less'"
      and subset: "\<forall>s t. less s t \<longrightarrow> less' s t"
      and min: "\<forall>t. ground t \<longrightarrow> less'\<^sup>=\<^sup>= t (Fun c [])"
    by (auto simp: less_def c_def)

  interpret ro: reduction_order less by fact
  interpret ro': gtotal_reduction_order less' by fact
 
  have R: "set R \<subseteq> ro'.less_set" using R and subset by (auto simp: less_def)
  
  have *: "{(x, y). less x y} \<subseteq> ro'.less_set" using subset by auto
  
  interpret oc: oc_spec less' "\<lambda>s t. check (less' s t) (shows '''')" by (unfold_locales) simp
  interpret gcr: gcr_spec x_var y_var "(\<lambda>_. check_joinable)" "(\<lambda>E _. check_instance E)" less'
    using oc.gcr_spec [OF ro'] .
  
  have "isOK (gcr.check_ECPs E R)" using ecps by (auto simp: check_ECPs_def)
  then have "GCR (ordstep ro'.less_set ?R)" by (rule gcr.check_ECPs [OF R])
  then show ?thesis using ro'.GCR_imp_CR_FGROUND [OF * ro fground min c F] by (simp add: less_def)
qed

definition "check_FGCR_run ro F E\<^sub>0 R\<^sub>0 E R steps = do {
  let check_ord = (\<lambda>s t. check (redord.less ro s t) (shows ''term pair cannot be oriented''));
  oc_ops.check_oriented check_ord R\<^sub>0;
  oc_ops.check_oc check_ord (E\<^sub>0, R\<^sub>0) (E, R) steps <+? (\<lambda>s. shows ''the oKB run could not be reconstructed\<newline>'' \<circ> shows_nl \<circ> s);
  check_FGCR ro F E R <+? (\<lambda>s. shows ''ground confluence could not be verified\<newline>'' \<circ> shows_nl \<circ> s)
}"

lemma check_FGCR_run:
  fixes F :: "('f::{show,compare_order} \<times> nat) list"
  fixes E :: "(('f, char list) term \<times> ('f, char list) term) list"
  fixes ro:: "('f, string) redord"
  assumes "isOK (check_FGCR_run (create_KBO_redord r F) F E\<^sub>0 R\<^sub>0 E R steps)"
  shows "CR (FGROUND (set F) (ordstep ({(s, t). redord.less (create_KBO_redord r F) s t}) (set (Rules E R)))) \<and>
         equivalent (set E\<^sub>0 \<union> set R\<^sub>0) (set E \<union> set R)"
proof -
  let ?F = "set F"
  let ?ER = "set (Rules E R)"

  define less :: "('f, string) term \<Rightarrow> ('f, string) term \<Rightarrow> bool" and c :: "'f"
    where
      "less = redord.less (create_KBO_redord r F)" and
      "c = redord.min_const (create_KBO_redord r F)"

  from assms have FGCR:"isOK(check_FGCR (create_KBO_redord r F) F E R)" by (auto simp: check_FGCR_run_def)

  have valid: "isOK (redord.valid (create_KBO_redord r F))"
    and ecps: "isOK (check_ECPs E R)"
    and F: "funas_trs ?ER \<subseteq> ?F"
    using FGCR by (auto simp: check_FGCR_def)

  from KBO_redord.valid [OF valid] obtain less'
    where ro: "reduction_order less"
      and c: "(c, 0) \<in> ?F"
      and fground: "\<forall>s t. fground ?F s \<and> fground ?F t \<longrightarrow> s = t \<or> less s t \<or> less t s"
      and ro': "gtotal_reduction_order less'"
      and subset: "\<forall>s t. less s t \<longrightarrow> less' s t"
      and min: "\<forall>t. ground t \<longrightarrow> less'\<^sup>=\<^sup>= t (Fun c [])"
    by (auto simp: less_def c_def)

  interpret ro: reduction_order less by fact
  interpret ro': gtotal_reduction_order less' by fact

  define check_ord where "check_ord = (\<lambda>s t. check (less s t) (shows ''term pair cannot be oriented''))"
  have check_ord: "\<And>s t. isOK (check_ord s t) \<longleftrightarrow> less s t" unfolding check_ord_def by auto
  define check_ord' where "check_ord' = (\<lambda>s t. check (less' s t) (shows ''term pair cannot be oriented''))"
  have check_ord': "\<And>s t. isOK (check_ord' s t) \<longleftrightarrow> less' s t" unfolding check_ord'_def by auto

  interpret oc: oc_spec less check_ord by (insert check_ord, unfold_locales) simp
  interpret oc': oc_spec less' check_ord' by (insert check_ord', unfold_locales) simp

  have oc_ok: "isOK (oc.check_oc (E\<^sub>0, R\<^sub>0) (E, R) steps)" and
       oriented_ok: "isOK (oc.check_oriented R\<^sub>0)"
    using assms[unfolded check_FGCR_run_def] unfolding check_ord_def less_def by force+

  note run = oc.check_oc[OF this]
  note oriented = oc.check_oriented[OF oriented_ok]
  with run oc.oKB'_rtrancl_less have "set R \<subseteq> ro.less_set" by auto
  with check_FGCR[OF FGCR] have CR:"CR (FGROUND ?F (ordstep {(s, t). less s t} ?ER))"
    using less_def by auto

  note equiv = oc.oKB_steps_conversion_permuted[OF run]
  with CR less_def show ?thesis
    unfolding equivalent_def conversion_def by (simp add: equiv esteps_is_conversion)
qed

text \<open>
We say that the system @{term E},@{term R} is a ground-complete rewrite system for the
equational system @{term E\<^sub>0} with respect to the reduction order @{term less} iff the
respective equational theories are the same ordered rewriting with @{term E},@{term R} is
confluent on ground terms.
\<close>

datatype 'f reduction_order_input =
    RPO_Input "'f status_prec_repr"
  | KBO_Input "'f prec_weight_repr"

definition
  ordered_completed_rewrite_system :: "('f::compare_order, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> 'f reduction_order_input \<Rightarrow> bool"
where
  "ordered_completed_rewrite_system E\<^sub>0 E R ord \<equiv>
   case ord of KBO_Input precw \<Rightarrow>
     let (p,w,w0,lcs,scf) = prec_weight_repr_to_prec_weight_funs precw in
     let less = \<lambda> s t. fst (kbo.kbo w w0 (scf_repr_to_scf scf) (\<lambda>f. f \<in> set lcs) (\<lambda> f g. fst (p f g)) (\<lambda> f g. snd (p f g)) s t) in
     equivalent E\<^sub>0 (E \<union> R) \<and> CR (FGROUND (funas_trs E\<^sub>0) (ordstep {(s, t). less s t} (E\<^sup>\<leftrightarrow> \<union> R)))"

end
