(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2013)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2013-2015)
License: LGPL (see file COPYING.LESSER)
*)
section \<open>Redex-Recording Proof Terms\<close>

theory Proof_Term
imports
  Trs
  First_Order_Terms.Matching
begin

text \<open>The type of proof terms.\<close>
type_synonym
  ('f, 'v) pterm = "(('f, 'v) rule + 'f, 'v) term"

text \<open>Multi-step rewriting with proof terms.\<close>
inductive
  mstepp :: "('f, 'v) trs \<Rightarrow> ('f, 'v) pterm \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  for R
where
  Var: "mstepp R (Var x) (Var x) (Var x)" |
  args: "\<And>f n As ss ts. \<lbrakk>length As = n; length ss = n; length ts = n;
    \<forall>i<n. mstepp R (As ! i) (ss ! i) (ts ! i)\<rbrakk> \<Longrightarrow>
    mstepp R (Fun (Inr f) As) (Fun f ss) (Fun f ts)" |
  rule: "\<And>l r n \<sigma> \<tau> As xs. \<lbrakk>length xs = n; length As = n;
    (l, r) \<in> R; xs = vars_term_list l; \<forall>i<n. mstepp R (As ! i) (\<sigma> (xs ! i)) (\<tau> (xs ! i))\<rbrakk> \<Longrightarrow>
    mstepp R (Fun (Inl (l, r)) As) (l \<cdot> \<sigma>) (r \<cdot> \<tau>)"

definition "mstep R A = {(x, y). mstepp R A x y}"

lemma mstepp_mpstep_eq [pred_set_conv]:
  "mstepp R A = (\<lambda>x y. (x, y) \<in> mstep R A)"
  by (simp add: mstep_def)

lemmas mstep_intros [intro?] = mstepp.intros [to_set]
lemmas mstep_induct [consumes 1, induct set: mstep] = mstepp.induct [to_set]
lemmas mstep_cases [consumes 1, cases set: mstep] = mstepp.cases [to_set]
lemmas mstep_simps = mstepp.simps [to_set]
lemmas mstep_Var = mstepp.Var [to_set]
lemmas mstep_args = mstepp.args [to_set]
lemmas mstep_rule = mstepp.rule [to_set]

lemma mstep_unique:
  assumes *: "\<And>l r. (l, r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
    and "(s, t) \<in> mstep R A" and "(s', t') \<in> mstep R A"
  shows "s = s' \<and> t = t'"
using assms(2-)
proof (induction arbitrary: s' t')
  case (Var x) then show ?case by (cases) simp
next
  case (args f n As ss ts)
  with args.prems show ?case by (cases) (simp, metis nth_equalityI)
next
  case (rule l r n \<sigma> \<tau> As xs)
  from rule.prems show ?case
  proof (cases)
    case (rule n' \<sigma>' \<tau>' xs')
    with rule.hyps have [simp]: "n' = n" "xs' = xs" by simp+
    from rule.IH and rule(7)
      have "\<forall>i<n. \<sigma> (xs ! i) = \<sigma>' (xs ! i) \<and> \<tau> (xs ! i) = \<tau>' (xs ! i)" by auto
    with \<open>xs = vars_term_list l\<close> and \<open>length xs = n\<close>
      have "\<forall>x\<in>set (vars_term_list l). \<sigma> x = \<sigma>' x \<and> \<tau> x = \<tau>' x" by (metis in_set_conv_nth)
    with * [OF \<open>(l, r) \<in> _\<close>] show ?thesis
      unfolding \<open>s' = _\<close> \<open>t' = _\<close> term_subst_eq_conv
      by auto
  qed
qed

lemma mstep_refl [simp]:
  "(t, t) \<in> mstep R (map_term Inr id t)"
  by (induct t) (auto intro: mstep_intros)

lemma mstep_ctxt:
  assumes "(s, t) \<in> mstep R A"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> mstep R ((map_ctxt Inr id C)\<langle>A\<rangle>)"
proof (induction C)
  case Hole with assms show ?case by simp
next
  case (More f ss C ts)
  let ?ss = "ss @ C\<langle>s\<rangle> # ts"
  let ?ts = "ss @ C\<langle>t\<rangle> # ts"
  let ?As = "map (map_term Inr id) ss @ (map_ctxt Inr id C)\<langle>A\<rangle> # map (map_term Inr id) ts"
  have "i = length ss \<Longrightarrow> (?ss ! i, ?ts ! i) \<in> mstep R (?As ! i)" for i
      using More.IH by (metis length_map nth_append_length)
  moreover
  have "i < length ss \<Longrightarrow> (?ss ! i, ?ts ! i) \<in> mstep R (?As ! i)" for i
    by (simp add: nth_append)
  moreover
  have "i < length ?ss \<Longrightarrow> i > length ss \<Longrightarrow> (?ss ! i, ?ts ! i) \<in> mstep R (?As ! i)" for i
    by (simp add: nth_append) 
  ultimately
  have "\<forall>i<length ?ss.  (?ss ! i, ?ts ! i) \<in> mstep R (?As ! i)"
    by (metis linorder_neqE_nat)
  from mstep_args [OF _ _ _ this, simplified]
  show ?case by (metis ctxt.simps(9) ctxt_apply_term.simps(2))
qed

lemma rstep_imp_mstep:
  assumes *: "\<And>l r. (l, r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
    and "(s, t) \<in> rstep R"
  shows "\<exists>A. (s, t) \<in> mstep R A"
using assms(2)
proof (induct)
  case (IH C \<sigma> l r)
  with * have "vars_term r \<subseteq> vars_term l" by simp
  let ?xs = "vars_term_list l"
  let ?As = "map (map_term Inr id) (map \<sigma> ?xs)"
  let ?n = "length ?xs"
  have "\<forall>i < ?n. (\<sigma> (?xs ! i), \<sigma> (?xs ! i)) \<in> mstep R (?As ! i)"
  proof (intro allI impI)
    fix i assume "i < ?n"
    then show "(\<sigma> (?xs ! i), \<sigma> (?xs ! i)) \<in> mstep R (?As ! i)"
      by simp
  qed
  from mstep_rule [OF _ _ \<open>(l, r) \<in> R\<close> _ this, simplified]
  have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> mstep R (Fun (Inl (l, r)) ?As)"
    by (simp add: IH.hyps mstep_rule)
  from mstep_ctxt [OF this] show ?case by blast
qed

lemma subst_rsteps_imp_rule_rsteps:
  assumes "\<forall>x\<in>vars_term l. (\<sigma> x, \<tau> x) \<in> (rstep R)\<^sup>*"
    and "vars_term r \<subseteq> vars_term l"
    and "(l, r) \<in> R"
  shows "(l \<cdot> \<sigma>, r \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
proof -
  from \<open>(l, r) \<in> R\<close> have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rstep R" by (metis rstep.rule rstep.subst)
  moreover have "(r \<cdot> \<sigma>, r \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
    by (rule subst_rsteps_imp_rsteps) (insert assms, auto)
  ultimately show ?thesis by auto
qed

lemma mstep_imp_rsteps:
  assumes *: "\<And>l r. (l, r) \<in> R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
    and "(s, t) \<in> mstep R A"
  shows "(s, t) \<in> (rstep R)\<^sup>*"
using assms(2-)
proof (induct)
  case (Var x) then show ?case by simp
next
  case (args f n As ss ts)
  then show ?case by (metis args_rsteps_imp_rsteps)
next
  case (rule l r n \<sigma> \<tau> As xs)
  then have "\<forall>x\<in>set xs. (\<sigma> x, \<tau> x) \<in> (rstep R)\<^sup>*"
    by (auto simp del: set_vars_term_list) (metis in_set_conv_nth)
  then have "\<forall>x\<in>vars_term l. (\<sigma> x, \<tau> x) \<in> (rstep R)\<^sup>*"
    using \<open>xs = vars_term_list l\<close> by simp
  then show ?case using * and \<open>(l, r) \<in> R\<close> by (metis subst_rsteps_imp_rule_rsteps)
qed

text \<open>Thus if @{term "{(x, y). \<exists>A. (x, y) \<in> mstep R A}"} has the diamond property,
then @{term "rstep R"} is confluent.\<close>

lemma Var_mstep:
  assumes *: "\<And>l r. (l, r) \<in> R \<Longrightarrow> \<not> is_Var l"
    and "(Var x, t) \<in> mstep R A"
  shows "t = Var x \<and> A = Var x"
  using assms(2-) by (cases) (insert *, simp, case_tac l, auto)

subsubsection \<open>A rewrite order on proof terms.\<close>

definition LESS1 :: "('f, 'v) trs \<Rightarrow> ('f, 'v) pterm rel" where
  "LESS1 R = {(Fun (Inl (l, r)) (map Var (vars_term_list l)), map_term Inr id l) | l r. (l, r) \<in> R}"

definition pleq :: "('f, 'v) trs \<Rightarrow> ('f, 'v) pterm rel" where
  "pleq R = (rstep (LESS1 R))\<^sup>*"

lemma pleq_refl [simp]:
  "(A, A) \<in> pleq R"
  by (simp add: pleq_def)

lemma pleq_trans:
  "(A, B) \<in> pleq R \<Longrightarrow> (B, C) \<in> pleq R \<Longrightarrow> (A, C) \<in> pleq R"
  by (simp add: pleq_def)

lemma pleq_ctxt:
  "(A, B) \<in> pleq R \<Longrightarrow> (C\<langle>A\<rangle>, C\<langle>B\<rangle>) \<in> pleq R"
  by (metis pleq_def rsteps_closed_ctxt)

lemma pleq_subst:
  "(A, B) \<in> pleq R \<Longrightarrow> (A \<cdot> \<sigma>, B \<cdot> \<sigma>) \<in> pleq R"
  by (metis pleq_def rsteps_closed_subst)

text \<open>A measure on proof terms (to show termination of @{term "rstep (LESS1 R)"}).\<close>
fun count_Inl :: "('f, 'v) pterm \<Rightarrow> nat" where
  "count_Inl (Var x) = 0" |
  "count_Inl (Fun (Inl (l, r)) As) = Suc (sum_list (map count_Inl As))" |
  "count_Inl (Fun (Inr f) As) = sum_list (map count_Inl As)"

lemma count_Inl_map_term_Inr [simp]:
  "count_Inl (map_term Inr id t) = 0"
  by (induct t) simp+

lemma LESS1_count_Inl:
  assumes "(l, r) \<in> LESS1 R"
  shows "count_Inl l > count_Inl r"
  using assms by (auto simp: LESS1_def)

lemma count_Inl_subst_apply_term [simp]:
  "count_Inl (t \<cdot> \<sigma>) =
    count_Inl t + sum_list (map count_Inl (map \<sigma> (vars_term_list t)))"
proof (induct t)
  case (Fun f ts)
  then have "sum_list (map count_Inl (map (\<lambda>x. x \<cdot> \<sigma>) ts)) =
    sum_list (map count_Inl ts) + sum_list (map count_Inl (map \<sigma> (concat (map vars_term_list ts))))"
    by (induct ts) auto
  then show ?case by (cases f, auto simp: vars_term_list.simps)
qed (simp add: vars_term_list.simps)

lemma var_map_term_Inr [simp]:
  "vars_term_list (map_term Inr id t) = vars_term_list t"
  by (induct t) (auto simp: o_def vars_term_list.simps, metis (lifting) map_eq_conv)

lemma Sum_var_count_Inl [simp]:
  fixes \<sigma> :: "'v \<Rightarrow> ('f, 'v) pterm"
  shows "(\<Sum>t\<leftarrow>ts. \<Sum>x\<leftarrow>var t. count_Inl (\<sigma> x)) = (\<Sum>x\<leftarrow>concat (map var ts). count_Inl (\<sigma> x))"
  by (induct ts) (auto simp: o_def)

lemma LESS1_count_Inl_subst:
  assumes "(l, r) \<in> LESS1 R"
  shows "count_Inl (l \<cdot> \<sigma>) > count_Inl (r \<cdot> \<sigma>)"
proof -
  from assms obtain l' r' where "(l', r') \<in> R"
    and "l = Fun (Inl (l', r')) (map Var (vars_term_list l'))"
    and "r = map_term Inr id l'" by (auto simp: LESS1_def)
  then show ?thesis by (induct l') (auto simp: vars_term_list.simps o_def)
qed

lemma count_Inl_ctxt_less:
  assumes "count_Inl s > count_Inl t"
  shows "count_Inl C\<langle>s\<rangle> > count_Inl C\<langle>t\<rangle>"
  using assms 
proof (induct C)
  case (More f bef C aft)
  then show ?case by (cases f, auto)
qed simp

lemma rstep_LESS1_count_Inl_less:
  assumes "(s, t) \<in> rstep (LESS1 R)"
  shows "count_Inl s > count_Inl t"
using assms
proof (induct)
  case (IH C \<sigma> l r)
  then have "count_Inl (l \<cdot> \<sigma>) > count_Inl (r \<cdot> \<sigma>)" by (rule LESS1_count_Inl_subst)
  then show ?case by (rule count_Inl_ctxt_less)
qed

lemma SN_rstep_LESS1:
  "SN (rstep (LESS1 R))"
  using SN_subset [OF SN_inv_image [OF SN_nat_gt, of count_Inl]]
    and rstep_LESS1_count_Inl_less [of _ _ R]
    by (auto simp: inv_image_def)

lemma pleq_antisym:
  assumes "(A, B) \<in> pleq R" and "(B, A) \<in> pleq R"
  shows "A = B"
proof (rule ccontr)
  assume "A \<noteq> B"
  with \<open>(A, B) \<in> pleq R\<close> have "(A, B) \<in> (rstep (LESS1 R))\<^sup>+"
    by (auto simp: pleq_def rtrancl_eq_or_trancl)
  with \<open>(B, A) \<in> pleq R\<close> have "(A, A) \<in> (rstep (LESS1 R))\<^sup>+"
    by (simp add: pleq_def)
  then have "\<not> SN ((rstep (LESS1 R))\<^sup>+)" by (auto simp: SN_defs)
  with SN_rstep_LESS1 [of R] show False
    by (metis SN_on_trancl_SN_on_conv)
qed

text \<open>@{const pleq} is a partial order on proof terms.\<close>
interpretation pterm_order:
  order "\<lambda>x y. (x, y) \<in> pleq R" "\<lambda>x y. (x, y) \<in> pleq R \<and> (y, x) \<notin> pleq R" for R
  by standard (auto intro: pleq_trans pleq_antisym)

text \<open>Auxiliary lemma in preparation of termination simp rule.\<close>
lemma match_vars_term_size:
  assumes "match (Fun f ss) t = Some \<sigma>"
    and *: "x \<in> vars_term t"
  shows "size (\<sigma> x) \<le> Suc (size_list size ss)"
proof -
  from match_sound [OF assms(1)] have **: "t \<cdot> \<sigma> = Fun f ss" unfolding matchers_def by auto
  show ?thesis
  proof (cases t)
    case (Var x) with * and ** show ?thesis by simp
  next
    case (Fun g ts)
    with ** have  "ss = map (\<lambda>x. x \<cdot> \<sigma>) ts" by simp+
    with * show ?thesis unfolding Fun
    proof (induct ts arbitrary: ss)
      case (Cons t ts)
      show ?case
      proof (cases "x \<in> vars_term t")
        case False
        with Cons show ?thesis
          by (auto) (metis add_Suc_right le_SucI trans_le_add2)
      next
        case True
        from vars_term_size [OF this, of \<sigma>] and Cons
          show ?thesis by simp
      qed
    qed simp
  qed
qed

lemma [termination_simp]:
  assumes "match (Fun f ss) l = Some \<sigma>"
    and *: "(s, t) \<in> set (zip (map \<sigma> (vars_term_list l)) ts)"
  shows "size s \<le> Suc (size_list size ss)"
proof -
  from * have "s \<in> set (map \<sigma> (vars_term_list l))" by (blast elim: in_set_zipE)
  then obtain x where [simp]: "s = \<sigma> x"
    and x: "x \<in> vars_term l" by (induct l) auto
  from match_vars_term_size [OF assms(1) x]
    show ?thesis by simp
qed

(* RT: do not know how to fix, did not investigate what "split" should be *)
(*
fun residual :: "('f, 'v) pterm \<Rightarrow> ('f, 'v) pterm \<Rightarrow> ('f, 'v) pterm" where
  "residual (Var x) (Var y) =
    (if x = y then (Var x) else undefined)" |
  "residual (Fun (Inr f) As) (Fun (Inr g) Bs) =
    (if f = g then Fun (Inr f) (map (split residual) (zip As Bs))
    else undefined)" |
  "residual (Fun (Inl (l, r)) As) (Fun (Inl (l', r')) Bs) =
    (if l = l' \<and> r = r' then map_term Inr id r \<cdot> mk_subst Var (zip (var l) (map (split residual) (zip As Bs)))
    else undefined)" |
  "residual (Fun (Inl (l, r)) As) (Fun (Inr g) Bs) =
    (case match (Fun (Inr g) Bs) (map_term Inr id l) of
      None \<Rightarrow> undefined
    | Some \<sigma> \<Rightarrow> Fun (Inl (l, r)) (map (split residual) (zip As (map \<sigma> (var l)))))" |
  "residual (Fun (Inr f) As) (Fun (Inl (l, r)) Bs) =
    (case match (Fun (Inr f) As) (map_term Inr id l) of
      None \<Rightarrow> undefined
    | Some \<sigma> \<Rightarrow> map_term Inr id r \<cdot> mk_subst Var (zip (var l) (map (split residual) (zip (map \<sigma> (var l)) Bs))))"
*)
end

